'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = plugin;

var _ImportRegistry = require('./ImportRegistry');

var _ImportRegistry2 = _interopRequireDefault(_ImportRegistry);

var _FlowTypeRegistry = require('./FlowTypeRegistry');

var _FlowTypeRegistry2 = _interopRequireDefault(_FlowTypeRegistry);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var KEY_MANIPULATION_TARGET = 'flow-onlyUpdateForKeys/manipulation-target';

function plugin(_ref) {
  var t = _ref.types;

  var componentRegistry = new _ImportRegistry2.default('react', ['Component', 'PureComponent']);
  var composeRegistry = new _ImportRegistry2.default('recompose', ['compose']);
  var onlyUpdateForKeysRegistry = new _ImportRegistry2.default('recompose', ['onlyUpdateForKeys']);
  var flowTypeRegistry = new _FlowTypeRegistry2.default();

  return {
    visitor: {
      Program: function Program(rootPath, _ref2) {
        var opts = _ref2.opts;

        componentRegistry.clear();
        composeRegistry.clear();
        onlyUpdateForKeysRegistry.clear();
        flowTypeRegistry.clear();

        rootPath.traverse({
          ImportDeclaration: function ImportDeclaration(path) {
            componentRegistry.scan(path);
            composeRegistry.scan(path);
            onlyUpdateForKeysRegistry.scan(path);
          },
          TypeAlias: function TypeAlias(path) {
            flowTypeRegistry.scan(path);
          },
          CallExpression: function CallExpression(path) {
            if (isManipulationTarget(path)) {
              var keys = getKeys(path);
              if (keys) {
                path.setData(KEY_MANIPULATION_TARGET, keys);
              }
            }
          }
        });
      },


      CallExpression: {
        exit: function exit(path) {
          var keys = path.getData(KEY_MANIPULATION_TARGET);
          if (keys) {
            path.pushContainer('arguments', [t.arrayExpression(keys.map(function (key) {
              return t.stringLiteral(key);
            }))]);
          }
        }
      }
    }
  };

  function isManipulationTarget(path) {
    return onlyUpdateForKeysRegistry.match(path.node.callee) && path.node.arguments.length === 0;
  }

  function getKeys(path) {
    var maybeCallCompose = path.findParent(function (parentPath) {
      return t.isCallExpression(parentPath.node);
    });
    if (!maybeCallCompose) {
      return null;
    }

    if (!composeRegistry.match(maybeCallCompose.node.callee)) {
      return null;
    }

    var maybeCallComposedFunction = maybeCallCompose.parentPath;
    if (!t.isCallExpression(maybeCallComposedFunction.node)) {
      return null;
    }

    var maybeComponent = maybeCallComposedFunction.node.arguments[0];
    if (t.isClassExpression(maybeComponent)) {
      return getKeysFromClassComponent(maybeComponent);
    }
    if (t.isFunctionExpression(maybeComponent) || t.isArrowFunctionExpression(maybeComponent)) {
      return getKeysFromFunctionalComponent(maybeComponent);
    }

    if (t.isIdentifier(maybeComponent)) {
      var referencedComponent = findReferencedComponent(maybeCallComposedFunction.scope, maybeComponent.name);
      if (!referencedComponent) {
        return null;
      }

      if (t.isClassDeclaration(referencedComponent) || t.isClassExpression(referencedComponent)) {
        return getKeysFromClassComponent(referencedComponent);
      }
      if (t.isFunctionDeclaration(referencedComponent) || t.isFunctionExpression(referencedComponent) || t.isArrowFunctionExpression(referencedComponent)) {
        return getKeysFromFunctionalComponent(referencedComponent);
      }
    }

    return null;
  }

  function findReferencedComponent(initialScope, name) {
    var scope = initialScope;
    while (scope) {
      if (scope.bindings[name]) {
        var referencedComponent = scope.bindings[name].path.node;
        if (t.isVariableDeclarator(referencedComponent)) {
          return referencedComponent.init;
        }
        return referencedComponent;
      }
      scope = scope.parent;
    }
    return null;
  }

  function getKeysFromClassComponent(component) {
    var superClass = component.superClass;
    if (!componentRegistry.match(superClass)) {
      return null;
    }

    var props = component.body.body.find(function (member) {
      return t.isClassProperty(member) && t.isIdentifier(member.key) && member.key.name === 'props';
    });
    if (!props) {
      return null;
    }

    return getKeysFromTypeAnnotation(props.typeAnnotation);
  }

  function getKeysFromFunctionalComponent(component) {
    if (component.params.length !== 1) {
      return null;
    }

    var props = component.params[0];
    if (!t.isIdentifier(props)) {
      return null;
    }

    return getKeysFromTypeAnnotation(props.typeAnnotation);
  }

  function getKeysFromTypeAnnotation(typeAnnotation) {
    if (!t.isTypeAnnotation(typeAnnotation)) {
      return null;
    }

    if (t.isObjectTypeAnnotation(typeAnnotation.typeAnnotation)) {
      var propsType = typeAnnotation.typeAnnotation;
      return propsType.properties.map(function (prop) {
        return prop.key.name;
      });
    }

    if (t.isGenericTypeAnnotation(typeAnnotation.typeAnnotation)) {
      var typeReference = typeAnnotation.typeAnnotation;
      var _propsType = flowTypeRegistry.getType(typeReference.id.name);
      if (!_propsType) {
        return null;
      }
      return _propsType.properties.map(function (prop) {
        return prop.key.name;
      });
    }

    return null;
  }
}